# -*- coding: utf-8 -*-
"""
Converts Discord library objects into generic objects that can be used by the AI

This is done to make it easier to swap out the Discord library for something else
in the future, and to make it easier to test the AI without having to mock the
Discord library.
"""


import base64
import re
import typing

import discord

from oobabot import fancy_logger
from oobabot import types

FORBIDDEN_CHARACTERS = r"[\n\r\t]"
FORBIDDEN_CHARACTERS_PATTERN = re.compile(FORBIDDEN_CHARACTERS)


def get_channel_name(channel: discord.abc.Messageable) -> str:
    if isinstance(channel, discord.Thread):
        return "thread #" + channel.name
    if isinstance(channel, discord.abc.GuildChannel):
        return "channel #" + channel.name
    if isinstance(channel, discord.DMChannel):
        return "-DM-"
    if isinstance(channel, discord.GroupChannel):
        return "-GROUP-DM-"
    return "-Unknown-"


def sanitize_string(raw_string: str) -> str:
    """
    Filter out any characters that would confuse the AI
    """
    return FORBIDDEN_CHARACTERS_PATTERN.sub(" ", raw_string)


def discord_message_to_generic_message(
    raw_message: discord.Message,
) -> typing.Union[types.GenericMessage, types.ChannelMessage, types.DirectMessage]:
    """
    Convert a discord message to a GenericMessage or subclass thereof
    """
    generic_args = {
        "author_id": raw_message.author.id,
        "author_name": sanitize_string(raw_message.author.display_name),
        "channel_id": raw_message.channel.id,
        "channel_name": get_channel_name(raw_message.channel),
        "message_id": raw_message.id,
        "body_text": sanitize_string(raw_message.content),
        "author_is_bot": raw_message.author.bot,
        "send_timestamp": raw_message.created_at.timestamp(),
        "reference_message_id": raw_message.reference.message_id
        if raw_message.reference
        else "",
    }
    if isinstance(raw_message.channel, discord.DMChannel):
        return types.DirectMessage(**generic_args)
    if isinstance(
        raw_message.channel, (discord.TextChannel, discord.GroupChannel, discord.Thread)
    ):
        return types.ChannelMessage(
            mentions=[mention.id for mention in raw_message.mentions],
            **generic_args,
        )
    fancy_logger.get().warning(
        f"Unknown channel type {type(raw_message.channel)}, "
        + f"unsolicited replies disabled.: {raw_message.channel}"
    )
    return types.GenericMessage(**generic_args)


def replace_mention_ids_with_names(
    generic_message: types.GenericMessage,
    fn_user_id_to_name: typing.Callable[["re.Match[str]"], str],
):
    """
    Replace user ID mentions with the user's chosen display
    name in the given guild (aka server)
    """
    # it looks like normal IDs are 18 digits.  But give it some
    # wiggle room in case things change in the future.
    # e.g.: <@009999999999999999>
    at_mention_pattern = r"<@(\d{16,20})>"
    while True:
        match = re.search(at_mention_pattern, generic_message.body_text)
        if not match:
            break
        generic_message.body_text = (
            generic_message.body_text[: match.start()]
            + fn_user_id_to_name(match)
            + generic_message.body_text[match.end() :]
        )


def dm_user_id_to_name(
    bot_user_id: int,
    bot_name: str,
) -> typing.Callable[["re.Match[str]"], str]:
    """
    Replace user ID mentions with the bot's name.  Used when
    we are in a DM with the bot.
    """
    if " " in bot_name:
        bot_name = f'"{bot_name}"'

    def _replace_user_id_mention(match: typing.Match[str]) -> str:
        user_id = int(match.group(1))
        print(f"bot_user_id={bot_user_id}, user_id={user_id}")
        if user_id == bot_user_id:
            return f"@{bot_name}"
        return match.group(0)

    return _replace_user_id_mention


def guild_user_id_to_name(
    guild: discord.Guild,
) -> typing.Callable[["re.Match[str]"], str]:
    def _replace_user_id_mention(match: typing.Match[str]) -> str:
        user_id = int(match.group(1))
        member = guild.get_member(user_id)
        if member is None:
            return match.group(0)
        display_name = member.display_name
        if " " in display_name:
            display_name = f'"{display_name}"'
        return f"@{display_name}"

    return _replace_user_id_mention


def get_intents() -> discord.Intents:
    intents = discord.Intents.default()
    intents.message_content = True
    intents.members = True
    return intents


async def test_discord_token(discord_token: str) -> bool:
    class _SimplestBot(discord.Client):
        async def on_ready(self):
            self.has_connected = True
            await self.close()

        def __init__(self):
            super().__init__(intents=get_intents())
            self.has_connected = False

    simplest_bot = _SimplestBot()
    try:
        await simplest_bot.start(discord_token, reconnect=False)
    except discord.LoginFailure:
        return False
    finally:
        await simplest_bot.close()
    return simplest_bot.has_connected


def get_user_id_from_token(discord_token: str) -> int:
    """
    Extract the bot's user ID from the discord token.
    """

    # turns out, the discord_token includes our client ID, so we can just
    # extract it from there.
    #
    # the discord token has this format:
    # AAAAAAAAAAAAAAAAAAAAAAAAAA.BBBBBB.CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
    #
    # where each section, A, B, and C, is independently a base64-encoded string.
    #
    # Section A encodes the bot's client ID, which is a 16-digit number.
    # the other sections aren't important here.
    token_parts = discord_token.split(".")
    token_part_a = token_parts[0]

    # the base64 decoder requires the string to be a multiple of 4 characters
    # long, so we need to add padding
    if len(token_part_a) % 4 != 0:
        token_part_a += "=" * (4 - len(token_part_a) % 4)

    return int(base64.b64decode(token_part_a).decode("utf-8"))


def generate_invite_url(ai_user_id: int) -> str:
    # we want to generate a URL like this:
    # https://discord.com/api/oauth2/authorize?client_id={client_id}&permissions={permissions}}&scope=bot
    #
    # where {client_id} is the bot's client ID, and {permissions} is the
    # permissions bit array with our desired permissions set.

    # for the permissions bit array, we can generate it with the library
    permissions = discord.Permissions(
        change_nickname=True,
        send_messages=True,
        create_public_threads=True,
        send_messages_in_threads=True,
        attach_files=True,
        read_message_history=True,
        read_messages=True,
        add_reactions=True,
    ).value

    return (
        "https://discord.com/api/oauth2/authorize?client_id="
        + f"{ai_user_id}&permissions={permissions}&scope=bot"
    )


def setup_logging(**kwargs: typing.Any):
    discord.utils.setup_logging(**kwargs)



# -*- coding: utf-8 -*-
"""
Purpose: Provides a base class for HTTP clients that limits the number
of connections to a single host to one, so that we don't overwhelm the
server.

This is a workaround for the fact that Oobabooga, at the time of writing,
could not support multiple pending requests at the same time without failing.
"""

import abc
import asyncio
import socket

import aiohttp


class OobaHttpClientError(Exception):
    """
    Purpose: Exception class for OobaHttpClient
    """


class SerializedHttpClient(abc.ABC):
    """
    Purpose: Limits the number of connections to a single host
    to one, so that we don't overwhelm the server.
    """

    HTTP_CLIENT_TIMEOUT_SECONDS: aiohttp.ClientTimeout = aiohttp.ClientTimeout(
        total=None,
        connect=None,
        sock_connect=5.0,
        sock_read=None,
    )

    @abc.abstractmethod
    async def _setup(self):
        # it's ok to raise an exception here, it will be caught
        ...

    async def setup(self):
        """
        Attempt to connect to the server.

        Returns:
            nothing, if the connection test was successful

        Raises:
            OobaHttpClientError, if the connection fails
        """
        try:
            await self._setup()
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientError,
            ConnectionRefusedError,
            socket.gaierror,
            asyncio.exceptions.TimeoutError,
        ) as err:
            raise OobaHttpClientError(
                f"Could not connect to {self.service_name} server: [{self.base_url}]"
            ) from err

    def __init__(self, service_name: str, base_url: str):
        self.service_name = service_name
        self.base_url = base_url
        self._session = None

    def _get_session(self) -> aiohttp.ClientSession:
        """
        Returns: the session, if it exists
        Raises: OobaHttpClientError if the session does not exist.
        """
        if not self._session:
            raise OobaHttpClientError("Session not initialized")
        return self._session

    async def __aenter__(self):
        connector = aiohttp.TCPConnector(limit_per_host=1)
        self._session = aiohttp.ClientSession(
            base_url=self.base_url,
            connector=connector,
            timeout=self.HTTP_CLIENT_TIMEOUT_SECONDS,
        )
        return self

    async def __aexit__(self, *_err):
        if self._session:
            await self._session.close()
        self._session = None

    def test_connection(self) -> None:
        async def try_setup():
            async with self:
                await self.setup()

        try:
            asyncio.run(try_setup())
        except AssertionError as err:
            # asyncio will throw an AssertionError if we try to run
            # with a base_url that has a path.  This is a user-supplied
            # value, so catching this is grody but necessary.
            raise OobaHttpClientError(
                f"Could not connect to {self.service_name} server: [{self.base_url}]"
            ) from err



# -*- coding: utf-8 -*-
"""
Implementation of the bot's slash commands.
"""
import typing

import discord

from oobabot import decide_to_respond
from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import persona
from oobabot import repetition_tracker
from oobabot import templates


class BotCommands:
    """
    Implementation of the bot's slash commands.
    """

    def __init__(
        self,
        decide_to_respond: decide_to_respond.DecideToRespond,
        repetition_tracker: repetition_tracker.RepetitionTracker,
        persona: persona.Persona,
        discord_settings: dict,
        template_store: templates.TemplateStore,
    ):
        self.decide_to_respond = decide_to_respond
        self.repetition_tracker = repetition_tracker
        self.persona = persona
        self.reply_in_thread = discord_settings["reply_in_thread"]
        self.template_store = template_store

    async def on_ready(self, client: discord.Client):
        """
        Register commands with Discord.
        """

        async def fail(
            interaction: discord.Interaction, reason: typing.Optional[str] = None
        ):
            msg = reason
            if msg is None:
                command = "command"
                if interaction.command is not None:
                    command = interaction.command.name
                fancy_logger.get().warning(
                    f"{command} called from an unexpected channel: "
                    + f"{interaction.channel_id}"
                )
                msg = f"{command} failed"

            await interaction.response.send_message(msg, ephemeral=True, silent=True)

        async def get_messageable(
            interaction: discord.Interaction,
        ) -> (
            typing.Optional[
                typing.Union[
                    discord.TextChannel,
                    discord.Thread,
                    discord.DMChannel,
                    discord.GroupChannel,
                ]
            ]
        ):
            if interaction.channel_id is not None:
                # find the current message in this channel
                # tell the Repetition Tracker to hide messages
                # before this message
                channel = await interaction.client.fetch_channel(interaction.channel_id)
                if channel is not None:
                    if isinstance(channel, discord.TextChannel):
                        return channel
                    if isinstance(channel, discord.Thread):
                        return channel
                    if isinstance(channel, discord.DMChannel):
                        return channel
                    if isinstance(channel, discord.GroupChannel):
                        return channel
            return None

        @discord.app_commands.command(
            name="say",
            description=f"Force {self.persona.ai_name} to say the provided message.",
        )
        @discord.app_commands.rename(text_to_send="message")
        @discord.app_commands.describe(
            text_to_send=f"Message to force {self.persona.ai_name} to say."
        )
        async def say(interaction: discord.Interaction, text_to_send: str):
            if interaction.channel_id is None:
                await fail(interaction)
                return

            # if reply_in_thread is True, we don't want our bot to
            # speak in guild channels, only threads and private messages
            if self.reply_in_thread:
                channel = await get_messageable(interaction)
                if channel is None or isinstance(channel, discord.TextChannel):
                    await fail(
                        interaction, f"{self.persona.ai_name} may only speak in threads"
                    )
                    return

            fancy_logger.get().debug(
                "/say called by user '%s' in channel #%d",
                interaction.user.name,
                interaction.channel_id,
            )
            # this will cause the bot to monitor the channel
            # and consider unsolicited responses
            self.decide_to_respond.log_mention(
                channel_id=interaction.channel_id,
                send_timestamp=interaction.created_at.timestamp(),
            )
            await interaction.response.send_message(
                text_to_send,
                suppress_embeds=True,
            )

        @discord.app_commands.command(
            name="lobotomize",
            description=f"Erase {self.persona.ai_name}'s memory of any message "
            + "before now in this channel.",
        )
        async def lobotomize(interaction: discord.Interaction):
            channel = await get_messageable(interaction)
            if channel is None:
                await fail(interaction)
                return

            # find the current message in this channel
            # tell the Repetition Tracker to hide messages
            # before this message
            async for message in channel.history(limit=1):
                channel_name = discord_utils.get_channel_name(channel)
                fancy_logger.get().info(
                    "/lobotomize called by user '%s' in #%s",
                    interaction.user.name,
                    channel_name,
                )
                self.repetition_tracker.hide_messages_before(
                    channel_id=channel.id,
                    message_id=message.id,
                )

            response = self.template_store.format(
                template_name=templates.Templates.COMMAND_LOBOTOMIZE_RESPONSE,
                format_args={
                    templates.TemplateToken.AI_NAME: self.persona.ai_name,
                    templates.TemplateToken.USER_NAME: interaction.user.name,
                },
            )
            await interaction.response.send_message(
                response,
                silent=True,
                suppress_embeds=True,
            )

        fancy_logger.get().debug(
            "Registering commands, sometimes this takes a while..."
        )

        tree = discord.app_commands.CommandTree(client)
        tree.add_command(lobotomize)
        tree.add_command(say)
        commands = await tree.sync(guild=None)
        for command in commands:
            fancy_logger.get().info(
                "Registered command: %s: %s", command.name, command.description
            )



# -*- coding: utf-8 -*-
"""
Client for the Ooba API.
Can provide the response by token or by sentence.
"""
import abc
import json
import re
import time
import typing

import aiohttp
import pysbd
import pysbd.utils

from oobabot import fancy_logger
from oobabot import http_client


class MessageSplitter(abc.ABC):
    """
    Split a response into separate messages.
    """

    # anything that can't be in a real response
    END_OF_INPUT = ""

    def __init__(self):
        self.printed_idx = 0
        self.full_response = ""

    def next(self, new_token: str) -> typing.Generator[str, None, None]:
        """
        Collects tokens into a single string, splits into messages
        by the subclass's logic, then yields each message as soon
        as it's found.

        Parameters:
            new_token: str, the next token to add to the string

        Returns:
            Generator[str, None, None], yields each sentence

        Note:
        When there is no longer any input, the caller must pass
        MessageSplitter.END_OF_INPUT to this function.  This
        function will then yield any remaining text, even if it
        doesn't look like a full sentence.
        """

        self.full_response += new_token
        unseen = self.full_response[self.printed_idx :]

        # if we've reached the end of input, yield it all,
        # even if we don't think it's a full sentence.
        if self.END_OF_INPUT == new_token:
            to_print = unseen.strip()
            if to_print:
                yield unseen
            self.printed_idx += len(unseen)
            return

        yield from self.partition(unseen)

    @abc.abstractmethod
    def partition(self, unseen: str) -> typing.Generator[str, None, None]:
        pass


class RegexSplitter(MessageSplitter):
    """
    Split a response into separate messages using a regex.
    """

    def __init__(self, regex: str):
        super().__init__()
        self.pattern = re.compile(regex)

    def partition(self, unseen: str) -> typing.Generator[str, None, None]:
        while True:
            match = self.pattern.match(unseen)
            if not match:
                break
            to_print = match.group(1)
            yield to_print
            self.printed_idx += match.end()
            unseen = self.full_response[self.printed_idx :]


class SentenceSplitter(MessageSplitter):
    """
    Split a response into separate messages using English
    sentence word breaks.
    """

    def __init__(self):
        super().__init__()
        self.segmenter = pysbd.Segmenter(language="en", clean=False, char_span=True)

    def partition(self, unseen: str) -> typing.Generator[str, None, None]:
        segments: typing.List[pysbd.utils.TextSpan] = self.segmenter.segment(
            unseen
        )  # type: ignore -- type is determined by char_span=True above

        # any remaining non-sentence things will be in the last element
        # of the list.  Don't print that yet.  At the very worst, we'll
        # print it when the END_OF_INPUT signal is reached.
        for sentence_w_char_spans in segments[:-1]:
            # sentence_w_char_spans is a class with the following fields:
            #  - sent: str, sentence text
            #  - start: start idx of 'sent', relative to original string
            #  - end: end idx of 'sent', relative to original string
            #
            # we want to remove the last '\n' if there is one.
            # we do want to include any other whitespace, though.

            to_print = sentence_w_char_spans.sent  # type: ignore
            if to_print.endswith("\n"):
                to_print = to_print[:-1]

            yield to_print

        # since we've printed all the previous segments,
        # the start of the last segment becomes the starting
        # point for the next round.
        if len(segments) > 0:
            self.printed_idx += segments[-1].start  # type: ignore


class OobaClient(http_client.SerializedHttpClient):
    """
    Client for the Ooba API.  Can provide the response by token or by sentence.
    """

    SERVICE_NAME = "Oobabooga"

    OOBABOOGA_STREAMING_URI_PATH: str = "/api/v1/stream"

    def __init__(
        self,
        settings: typing.Dict[str, typing.Any],
    ):
        super().__init__(self.SERVICE_NAME, settings["base_url"])
        self.total_response_tokens = 0
        self.message_regex = settings["message_regex"]
        self.request_params = settings["request_params"]
        self.log_all_the_things = settings["log_all_the_things"]

        if self.message_regex:
            self.fn_new_splitter = lambda: RegexSplitter(self.message_regex)
        else:
            self.fn_new_splitter = SentenceSplitter

    def on_ready(self):
        """
        Called when the client is ready to start.
        Used to log our configuration.
        """
        if self.message_regex:
            fancy_logger.get().debug(
                "Ooba Client: Splitting responses into messages " + "with: %s",
                self.message_regex,
            )
        else:
            fancy_logger.get().debug(
                "Ooba Client: Splitting responses into messages "
                + "by English sentence.",
            )

    async def _setup(self):
        async with self._get_session().ws_connect(self.OOBABOOGA_STREAMING_URI_PATH):
            return

    def get_stopping_strings(self) -> typing.List[str]:
        """
        Returns a list of strings that indicate the end of a response.
        Taken from the yaml `stopping_strings` within our
        response_params.
        """
        return self.request_params.get("stopping_strings", [])

    async def request_by_message(self, prompt: str) -> typing.AsyncIterator[str]:
        """
        Yields individual messages from the response as it arrives.
        These can be split by a regex or by sentence.
        """
        splitter = self.fn_new_splitter()
        async for new_token in self.request_by_token(prompt):
            for sentence in splitter.next(new_token):
                # remove "### Assistant: " from strings
                if sentence.startswith("### Assistant: "):
                    sentence = sentence[len("### Assistant: ") :]
                yield sentence

    async def request_as_string(self, prompt: str) -> str:
        """
        Yields the entire response as a single string.
        """
        return "".join([token async for token in self.request_by_token(prompt)])

    async def request_as_grouped_tokens(
        self,
        prompt: str,
        interval: float = 0.2,
    ) -> typing.AsyncIterator[str]:
        """
        Yields the response as a series of tokens, grouped by time.
        """

        last_response = time.perf_counter()
        tokens = ""
        async for token in self.request_by_token(prompt):
            if token == SentenceSplitter.END_OF_INPUT:
                if tokens:
                    yield tokens
                break
            tokens += token
            now = time.perf_counter()
            if now < (last_response + interval):
                continue
            yield tokens
            tokens = ""
            last_response = time.perf_counter()

    async def request_by_token(self, prompt: str) -> typing.AsyncIterator[str]:
        """
        Yields each token of the response as it arrives.
        """

        request: dict[
            str, typing.Union[bool, float, int, str, typing.List[typing.Any]]
        ] = {
            "prompt": prompt,
        }
        request.update(self.request_params)

        async with self._get_session().ws_connect(
            self.OOBABOOGA_STREAMING_URI_PATH
        ) as websocket:
            await websocket.send_json(request)
            if self.log_all_the_things:
                print(f"Sent request:\n{json.dumps(request, indent=1)}")
                print(f"Prompt:\n{request['prompt']}")

            async for msg in websocket:
                # we expect a series of text messages in JSON encoding,
                # like this:
                #
                # {"event": "text_stream", "message_num": 0, "text": ""}
                # {"event": "text_stream", "message_num": 1, "text": "Oh"}
                # {"event": "text_stream", "message_num": 2, "text": ","}
                # {"event": "text_stream", "message_num": 3, "text": " okay"}
                # {"event": "text_stream", "message_num": 4, "text": "."}
                # {"event": "stream_end", "message_num": 5}
                # get_logger().debug(f"Received message: {msg}")
                if msg.type == aiohttp.WSMsgType.TEXT:
                    # bdata = typing.cast(bytes, msg.data)
                    # get_logger().debug(f"Received data: {bdata}")

                    incoming_data = msg.json()
                    if "text_stream" == incoming_data["event"]:
                        self.total_response_tokens += 1
                        text = incoming_data["text"]
                        if text != SentenceSplitter.END_OF_INPUT:
                            if self.log_all_the_things:
                                print(text, end="", flush=True)

                            yield text

                    elif "stream_end" == incoming_data["event"]:
                        # Make sure any unprinted text is flushed.
                        if self.log_all_the_things:
                            print("", flush=True)
                        yield SentenceSplitter.END_OF_INPUT
                        return

                    else:
                        fancy_logger.get().warning(
                            "Unexpected event: %s", incoming_data
                        )

                elif msg.type == aiohttp.WSMsgType.ERROR:
                    fancy_logger.get().error(
                        "WebSocket connection closed with error: %s", msg
                    )
                    raise http_client.OobaHttpClientError(
                        f"WebSocket connection closed with error {msg}"
                    )
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    fancy_logger.get().info(
                        "WebSocket connection closed normally: %s", msg
                    )
                    return



# -*- coding: utf-8 -*-
"""
Client for generating images using the AUTOMATIC1111
API.  Takes a prompt and returns image binary data in PNG format.
"""

import asyncio
import base64
import time
import typing

import aiohttp

from oobabot import fancy_logger
from oobabot import http_client

# todo: response stats for SD client


class StableDiffusionClient(http_client.SerializedHttpClient):
    """
    Purpose: Client for generating images using the AUTOMATIC1111
    API.  Takes a prompt and returns image binary data in PNG format.
    """

    SERVICE_NAME = "Stable Diffusion"
    STABLE_DIFFUSION_API_URI_PATH: str = "/sdapi/v1/"

    API_COMMAND_URLS = {
        "get_samplers": STABLE_DIFFUSION_API_URI_PATH + "samplers",
        # get_samplers: GET only
        #   returns a list of samplers which we can use
        #   in text2img
        "options": STABLE_DIFFUSION_API_URI_PATH + "options",
        # options: GET and POST
        #   retrieves (GET) and sets (POST) global options
        #   for the server.  This is how we set the checkpoint
        #   and also a few privacy-related fields.
        "progress": STABLE_DIFFUSION_API_URI_PATH + "progress",
        # progress: GET only
        #   returns the progress of the current image generation
        "txt2img": STABLE_DIFFUSION_API_URI_PATH + "txt2img",
        # txt2img: POST only
        #   takes a prompt, generates an image and returns it
    }

    def __init__(
        self,
        settings: typing.Dict[str, typing.Any],
    ):
        super().__init__(self.SERVICE_NAME, settings["stable_diffusion_url"])

        self.extra_prompt_text = settings["extra_prompt_text"]
        self.request_params = settings["request_params"]

        # when we're in a "age restricted" channel, we'll swap
        # the "negative_prompt" in the request_params with this
        # value.  Otherwise we'll use the one already in
        # request_params["negative_prompt"]
        self.negative_prompt_nsfw = self.request_params.pop("negative_prompt_nsfw", "")

    DEFAULT_OPTIONS = {
        #
        # - "enable_pnginfo"
        #   by default, this will store the parameters used to generate the image.
        #   For instance:
        #         Parameters: zombie taylor swift pinup movie posterSteps: 30,
        #           Sampler: DPM++ 2M Karras, CFG scale: 7, Seed: 1317257172,
        #           Size: 768x768, Model hash: 9aba26abdf, Model: Deliberate-2.0-15236
        #   We want to disable this, for a few reasons:
        #   - it's a privacy concern, since it's not clear to the user what
        #     is happening
        #   - it will include our negative prompts, which are actually bad
        #     phrases that represent content we DON'T want to generate.  If
        #     we upload these same strings to Discord, they might mistakenly
        #     flag the content as violating their TOS.  Even though the only
        #     reason those strings are used is to prevent that content.
        "enable_pnginfo": False,
        #
        # - "do_not_add_watermark"
        #    Similar to the above, we don't want to add a watermark to the
        #    generated image, just since it looks bad.  Also, it can leak
        #    the version of the SD software we are using, which isn't a good
        #    idea for security reasons.
        "do_not_add_watermark": True,
        #
        # - "samples_format"
        #   this is the format used for all image generation.
        "samples_format": "png",
    }

    async def set_options(self):
        url = self.API_COMMAND_URLS["options"]

        # first, get the current options.  We do this for two
        # reasons;
        #  - show the user exactly what we changed, if anything
        #  - some versions of the API don't support the
        #    "do_not_add_watermark" option, so we need to
        #    check if it's there before we try to set it
        #
        current_options = None
        async with self._get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            current_options = await response.json()

        options_to_set = {}
        for option_name, option_value in self.DEFAULT_OPTIONS.items():
            if option_name not in current_options:
                continue
            if option_value == current_options[option_name]:
                continue
            options_to_set[option_name] = option_value
            fancy_logger.get().info(
                "Stable Diffusion:  changing option '%s' from to '%s' to '%s'",
                option_name,
                current_options[option_name],
                option_value,
            )

        if not options_to_set:
            fancy_logger.get().debug(
                "Stable Diffusion: Options are already set correctly, no changes made."
            )
            return

        async with self._get_session().post(url, json=options_to_set) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            await response.json()

    async def get_samplers(self) -> typing.List[str]:
        url = self.API_COMMAND_URLS["get_samplers"]
        async with self._get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            response = await response.json()
            samplers = [str(sampler["name"]) for sampler in response]
            return samplers

    def generate_image(
        self,
        prompt: str,
        is_channel_nsfw: bool = False,
    ) -> "asyncio.Task[bytes]":
        """
        Generate an image from a prompt.
        Args:
            prompt: The prompt to generate an image from.
            is_channel_nsfw: Whether the channel is NSFW.
            this will change the negative prompt.
        Returns:
            The image as bytes.
        Raises:
            OobaHttpClientError, if the request fails.
        """
        request = self.request_params.copy()
        request["prompt"] = prompt
        if is_channel_nsfw:
            request["negative_prompt"] = self.negative_prompt_nsfw

        if self.extra_prompt_text:
            request["prompt"] += ", " + self.extra_prompt_text

        async def do_post() -> bytes:
            fancy_logger.get().debug(
                "Stable Diffusion: Image request (nsfw: %r): %s",
                is_channel_nsfw,
                request["prompt"],
            )
            start_time = time.time()

            async with self._get_session().post(
                self.API_COMMAND_URLS["txt2img"],
                json=request,
            ) as response:
                if response.status != 200:
                    raise http_client.OobaHttpClientError(response)
                duration = time.time() - start_time
                json_body = await response.json()
                image_bytes = base64.b64decode(json_body["images"][0])
                fancy_logger.get().debug(
                    "Stable Diffusion: Image generated, %d bytes in %.2f seconds",
                    len(image_bytes),
                    duration,
                )
                return image_bytes

        # works around aiohttp being bad
        async def do_post_with_retry() -> bytes:
            tries = 0
            while True:
                try:
                    return await do_post()
                except (aiohttp.ClientError, aiohttp.ClientOSError) as err:
                    retry = False
                    if err.__cause__ is ConnectionResetError:
                        retry = True
                    if isinstance(err, aiohttp.ClientOSError) and 104 == err.errno:
                        retry = True
                    if tries > 2:
                        retry = False
                    if not retry:
                        raise

                    fancy_logger.get().warning(
                        "Stable Diffusion: Connection reset error: %s, "
                        + "retrying in 1 second",
                        err,
                    )
                    await asyncio.sleep(1)
                    tries += 1

        return asyncio.create_task(do_post_with_retry())

    async def verify_sampler_available(self):
        """
        Checks that the requested sampler is available on the server.
        If it isn't, logs a warning and sets the sampler to the default.
        """
        samplers = await self.get_samplers()

        desired_sampler = self.request_params.get("sampler")
        if not desired_sampler:
            fancy_logger.get().debug(
                "Stable Diffusion: Using default sampler on SD server"
            )
            return

        if desired_sampler in samplers:
            fancy_logger.get().debug(
                "Stable Diffusion: Using desired sampler '%s'", desired_sampler
            )
            return

        fancy_logger.get().warning(
            "Stable Diffusion: Desired sampler '%s' not available",
            desired_sampler,
        )
        fancy_logger.get().info(
            "Stable Diffusion: Available samplers: %s", ", ".join(samplers)
        )
        self.request_params["sampler"] = ""

    async def _setup(self):
        await self.verify_sampler_available()
        await self.set_options()
        fancy_logger.get().debug(
            "Stable Diffusion: Using negative prompt: %s...",
            str(self.request_params.get("negative_prompt", ""))[:20],
        )
        if self.extra_prompt_text:
            fancy_logger.get().debug(
                "Stable Diffusion: will append to every prompt: '%s'",
                self.extra_prompt_text,
            )
